#include <amsculib2/amsculib2.hpp>

namespace amscuda
{

__host__ __device__ cuvect4::cuvect4()
{
    x = 0.0; y = 0.0; z = 0.0; w = 0.0;
    return;
}

__host__ __device__ cuvect4::~cuvect4()
{
    x = 0.0; y = 0.0; z = 0.0; w = 0.0;
    return;
}

__host__ __device__ cuvect4::cuvect4(double _x, double _y, double _z, double _w)
{
    x = _x; y = _y; z = _z; w = _w;
    return;
}
        
__host__ __device__ double& cuvect4::operator[](const int I)
{
    if(I==0) return x;
    else if(I==1) return y;
    else if(I==2) return z;
    else if(I==3) return w;
    return x;
}

__host__ __device__ const double& cuvect4::operator[](const int I) const
{
    if(I==0) return x;
    else if(I==1) return y;
    else if(I==2) return z;
    else if(I==3) return w;
    return x;
}

__host__ __device__ cuvect4 cuvect4::operator+(cuvect4 lhs)
{
    cuvect4 ret;
    ret.x = this->x + lhs.x;
    ret.y = this->y + lhs.y;
    ret.z = this->z + lhs.z;
    ret.w = this->w + lhs.w;
    return ret;
}

__host__ __device__ cuvect4 cuvect4::operator-(cuvect4 lhs)
{
    cuvect4 ret;
    ret.x = this->x - lhs.x;
    ret.y = this->y - lhs.y;
    ret.z = this->z - lhs.z;
    ret.w = this->w - lhs.w;
    return ret;
}

__host__ __device__ cuvect4 cuvect4::operator*(double lhs)
{
    cuvect4 ret;
    ret.x = this->x*lhs;
    ret.y = this->y*lhs;
    ret.z = this->z*lhs;
    ret.w = this->w*lhs;
    return ret;
}

__host__ __device__ cuvect4 cuvect4::operator/(double lhs)
{
    cuvect4 ret;
    ret.x = this->x/lhs;
    ret.y = this->y/lhs;
    ret.z = this->z/lhs;
    ret.w = this->w/lhs;
    return ret;
}

__host__ __device__ cumat4::cumat4()
{
    int I;
    for(I=0;I<16;I++)
    {
        dat[I] = 0.0;
    }
    return;
}

__host__ __device__ cumat4::~cumat4()
{
    int I;
    for(I=0;I<16;I++)
    {
        dat[I] = 0.0;
    }
    return;
}

__host__ __device__ double& cumat4::operator[](const int I)
{
    return dat[I];
}

__host__ __device__ double& cumat4::operator()(const int I, const int J)
{
    return dat[I+4*J];
}

__host__ __device__ double& cumat4::at(const int I, const int J)
{
    return dat[I+4*J];
}

__host__ __device__ cumat4 cumat4::operator+(cumat4 lhs)
{
    cumat4 ret;
    int I;
    for(I=0;I<16;I++)
    {
        ret.dat[I] = this->dat[I] + lhs.dat[I];
    }
    return ret;
}

__host__ __device__ cumat4 cumat4::operator-(cumat4 lhs)
{
    cumat4 ret;
    int I;
    for(I=0;I<16;I++)
    {
        ret.dat[I] = this->dat[I] - lhs.dat[I];
    }
    return ret;
}

__host__ __device__ cumat4 cumat4::operator*(double lhs)
{
    cumat4 ret;
    int I;
    for(I=0;I<16;I++)
    {
        ret.dat[I] = this->dat[I]*lhs;
    }
    return ret;
}

__host__ __device__ cumat4 cumat4::operator/(double lhs)
{
    cumat4 ret;
    int I;
    for(I=0;I<16;I++)
    {
        ret.dat[I] = this->dat[I]/lhs;
    }
    return ret;
}

__host__ __device__ cuvect4 cumat4::operator*(cuvect4 lhs)
{
    cuvect4 ret = cuvect4(0.0,0.0,0.0,0.0);
    int I,J;

    for(I=0;I<4;I++)
    {
        for(J=0;J<4;J++)
        {
            ret[I] = ret[I] + this->at(I,J)*lhs[J];
        }
    }

    return ret;
}

__host__ __device__ cumat4 cumat4::operator*(cumat4 lhs)
{
    cumat4 ret;
    int I,J,K;
    for(I=0;I<4;I++)
    {
        for(J=0;J<4;J++)
        {
            ret(I,J) = 0;
            for(K=0;K<4;K++)
            {
                ret(I,J) = ret(I,J) + this->at(I,K) * lhs(K,J);
            }
        }
    }
    return ret;
}

__host__ __device__ cumat4 cumat4::transpose()
{
    cumat4 q;
    int I,J;
    for(I=0;I<4;I++)
    {
        for(J=0;J<4;J++)
        {
            q(I,J) = this->at(J,I);
        }
    }
    return q;
}

__host__ __device__ double cumat4::det()
{
    double a00,a01,a02,a03;
    double a10,a11,a12,a13;
    double a20,a21,a22,a23;
    double a30,a31,a32,a33;
    double det;

    a00 = this->at(0,0);
    a01 = this->at(0,1);
    a02 = this->at(0,2);
    a03 = this->at(0,3);
    a10 = this->at(1,0);
    a11 = this->at(1,1);
    a12 = this->at(1,2);
    a13 = this->at(1,3);
    a20 = this->at(2,0);
    a21 = this->at(2,1);
    a22 = this->at(2,2);
    a23 = this->at(2,3);
    a30 = this->at(3,0);
    a31 = this->at(3,1);
    a32 = this->at(3,2);
    a33 = this->at(3,3);

    det = a03*a12*a21*a30 -
        a02*a13*a21*a30 -
        a03*a11*a22*a30 +
        a01*a13*a22*a30 +
        a02*a11*a23*a30 -
        a01*a12*a23*a30 -
        a03*a12*a20*a31 +
        a02*a13*a20*a31 +
        a03*a10*a22*a31 -
        a00*a13*a22*a31 -
        a02*a10*a23*a31 +
        a00*a12*a23*a31 +
        a03*a11*a20*a32 -
        a01*a13*a20*a32 -
        a03*a10*a21*a32 +
        a00*a13*a21*a32 +
        a01*a10*a23*a32 -
        a00*a11*a23*a32 -
        a02*a11*a20*a33 +
        a01*a12*a20*a33 +
        a02*a10*a21*a33 -
        a00*a12*a21*a33 -
        a01*a10*a22*a33 +
        a00*a11*a22*a33;

    return det;
}

__host__ __device__ cumat4 minverse(cumat4 ma)
{
  cumat4 mb;

  double a00,a01,a02,a03;
  double a10,a11,a12,a13;
  double a20,a21,a22,a23;
  double a30,a31,a32,a33;

  double b00,b01,b02,b03;
  double b10,b11,b12,b13;
  double b20,b21,b22,b23;
  double b30,b31,b32,b33;

  double det = 0.0;

  a00 = ma.at(0,0);
  a01 = ma.at(0,1);
  a02 = ma.at(0,2);
  a03 = ma.at(0,3);
  a10 = ma.at(1,0);
  a11 = ma.at(1,1);
  a12 = ma.at(1,2);
  a13 = ma.at(1,3);
  a20 = ma.at(2,0);
  a21 = ma.at(2,1);
  a22 = ma.at(2,2);
  a23 = ma.at(2,3);
  a30 = ma.at(3,0);
  a31 = ma.at(3,1);
  a32 = ma.at(3,2);
  a33 = ma.at(3,3);

  det = a03*a12*a21*a30 -
        a02*a13*a21*a30 -
        a03*a11*a22*a30 +
        a01*a13*a22*a30 +
        a02*a11*a23*a30 -
        a01*a12*a23*a30 -
        a03*a12*a20*a31 +
        a02*a13*a20*a31 +
        a03*a10*a22*a31 -
        a00*a13*a22*a31 -
        a02*a10*a23*a31 +
        a00*a12*a23*a31 +
        a03*a11*a20*a32 -
        a01*a13*a20*a32 -
        a03*a10*a21*a32 +
        a00*a13*a21*a32 +
        a01*a10*a23*a32 -
        a00*a11*a23*a32 -
        a02*a11*a20*a33 +
        a01*a12*a20*a33 +
        a02*a10*a21*a33 -
        a00*a12*a21*a33 -
        a01*a10*a22*a33 +
        a00*a11*a22*a33;

  if(det*det>1.0E-30)
  {
    b00 = -a13*a22*a31 + a12*a23*a31 + a13*a21*a32 - a11*a23*a32 - a12*a21*a33 + a11*a22*a33;
    b01 =  a03*a22*a31 - a02*a23*a31 - a03*a21*a32 + a01*a23*a32 + a02*a21*a33 - a01*a22*a33;
    b02 = -a03*a12*a31 + a02*a13*a31 + a03*a11*a32 - a01*a13*a32 - a02*a11*a33 + a01*a12*a33;
    b03 =  a03*a12*a21 - a02*a13*a21 - a03*a11*a22 + a01*a13*a22 + a02*a11*a23 - a01*a12*a23;
    b10 =  a13*a22*a30 - a12*a23*a30 - a13*a20*a32 + a10*a23*a32 + a12*a20*a33 - a10*a22*a33;
    b11 = -a03*a22*a30 + a02*a23*a30 + a03*a20*a32 - a00*a23*a32 - a02*a20*a33 + a00*a22*a33;
    b12 =  a03*a12*a30 - a02*a13*a30 - a03*a10*a32 + a00*a13*a32 + a02*a10*a33 - a00*a12*a33;
    b13 = -a03*a12*a20 + a02*a13*a20 + a03*a10*a22 - a00*a13*a22 - a02*a10*a23 + a00*a12*a23;
    b20 = -a13*a21*a30 + a11*a23*a30 + a13*a20*a31 - a10*a23*a31 - a11*a20*a33 + a10*a21*a33;
    b21 =  a03*a21*a30 - a01*a23*a30 - a03*a20*a31 + a00*a23*a31 + a01*a20*a33 - a00*a21*a33;
    b22 = -a03*a11*a30 + a01*a13*a30 + a03*a10*a31 - a00*a13*a31 - a01*a10*a33 + a00*a11*a33;
    b23 =  a03*a11*a20 - a01*a13*a20 - a03*a10*a21 + a00*a13*a21 + a01*a10*a23 - a00*a11*a23;
    b30 =  a12*a21*a30 - a11*a22*a30 - a12*a20*a31 + a10*a22*a31 + a11*a20*a32 - a10*a21*a32;
    b31 = -a02*a21*a30 + a01*a22*a30 + a02*a20*a31 - a00*a22*a31 - a01*a20*a32 + a00*a21*a32;
    b32 =  a02*a11*a30 - a01*a12*a30 - a02*a10*a31 + a00*a12*a31 + a01*a10*a32 - a00*a11*a32;
    b33 = -a02*a11*a20 + a01*a12*a20 + a02*a10*a21 - a00*a12*a21 - a01*a10*a22 + a00*a11*a22;
    b00 = b00/det;
    b01 = b01/det;
    b02 = b02/det;
    b03 = b03/det;
    b10 = b10/det;
    b11 = b11/det;
    b12 = b12/det;
    b13 = b13/det;
    b20 = b20/det;
    b21 = b21/det;
    b22 = b22/det;
    b23 = b23/det;
    b30 = b30/det;
    b31 = b31/det;
    b32 = b32/det;
    b33 = b33/det;
    mb.at(0,0) = b00;
    mb.at(0,1) = b01;
    mb.at(0,2) = b02;
    mb.at(0,3) = b03;
    mb.at(1,0) = b10;
    mb.at(1,1) = b11;
    mb.at(1,2) = b12;
    mb.at(1,3) = b13;
    mb.at(2,0) = b20;
    mb.at(2,1) = b21;
    mb.at(2,2) = b22;
    mb.at(2,3) = b23;
    mb.at(3,0) = b30;
    mb.at(3,1) = b31;
    mb.at(3,2) = b32;
    mb.at(3,3) = b33;
  }
  //this was STUPID. Gaah. Computer algebra system saves the day? I'd be surprised if this didn't end up *slower* than gaussian elimination. Don't do this again!
  return mb;
}

__host__ __device__ cumat4 cumat4::inverse()
{
    return minverse(*this);
}


__host__ __device__ double cuvect4_dot(cuvect4 a, cuvect4 b)
{
    double ret = 0.0;
    ret = a.x*b.x + a.y*b.y + a.z*b.z + a.w*b.w;
    return ret;
}

__host__ __device__ double cuvect4_norm(cuvect4 a)
{
    double ret = 0.0;
    ret = ::sqrt(cuvect4_dot(a,a));
    return ret;
}

__host__ __device__ cuvect4 cuvect4_normalize(cuvect4 a)
{
    cuvect4 ret = cuvect4(0.0f,0.0f,0.0f,0.0f);
    double nrm = cuvect4_norm(a);
    if(nrm>0.0)
        ret = a/nrm;
    return ret;
}

__host__ __device__ cuvect4 cuvect4_proj(cuvect4 a, cuvect4 b)
{
    cuvect4 ret;
    cuvect4 bn = cuvect4_normalize(b);
    double d = cuvect4_dot(a,bn);
    ret = bn*d;
    return ret;
}


};